# -*- coding: utf-8 -*-
"""
obspy.io.ah - AH (Ad Hoc) read support for ObsPy
================================================
This module provides read support for the AH (Ad Hoc) waveform format
introduced by the Lamont-Doherty Geological Observatory.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

"""

if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)
