# -*- coding: utf-8 -*-
"""
obspy.io.css - CSS read support for ObsPy
=========================================
This module provides read support for CSS waveform data.

:copyright:
    The ObsPy Development Team (devs@obspy.org), Marcus Walther
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

"""

if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)
