# -*- coding: utf-8 -*-
"""
obspy.io.cnv - CNV file format support for ObsPy
================================================

This module provides write support for CNV event summary file
format as used by VELEST program.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""

if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)
